<?php
// $Id: bitcache.inc,v 1.4 2008/05/09 11:15:59 arto Exp $
/**
 * bitcache.inc - Distributed content-addressable storage system.
 *
 * @author Arto Bendiken <http://bendiken.net/>
 * @copyright Copyright (c) 2007-2008 Arto Bendiken. All rights reserved.
 * @license GPL <http://creativecommons.org/licenses/GPL/2.0/>
 * @package bitcache.module
 */

//////////////////////////////////////////////////////////////////////////////

define('BITCACHE_ID_FORMAT', '/^[a-z0-9]{40}$/');
define('BITCACHE_LAST_MODIFIED', TRUE);
define('BITCACHE_POST_MAX_SIZE', (int)substr(ini_get('post_max_size'), 0, -1) * pow(2, array_search(strtolower(substr(ini_get('post_max_size'), -1)), array(1 => '', 10 => 'k', 20 => 'm', 30 => 'g'))));

//////////////////////////////////////////////////////////////////////////////
// Bitstream implementation

/**
 * Implements a wrapper for accessing individual bitstreams.
 */
class Bitcache_Stream {
  public $path, $size, $type;
  public $compressed = FALSE;
  public $encrypted = FALSE;

  function __construct($path) {
    $this->path = $path;
    $this->size = filesize($path);
    $this->type = self::mime_type($path);
  }

  function open($mode = 'rb') {
    return $this->stream = fopen($this->path, $mode);
  }

  function close() {
    if (!empty($this->stream)) {
      fclose($this->stream);
      unset($this->stream);
    }
  }

  function mime_type($filename) {
    if (extension_loaded('fileinfo') && ($finfo = finfo_open(FILEINFO_MIME))) {
      $type = finfo_file($finfo, $filename);
      finfo_close($finfo);
      return $type;
    }
    else if (function_exists('mime_content_type')) {
      return mime_content_type($filename);
    }
    else if (file_exists($filename)) {
      // Attempts to detect a file's MIME type using the Unix `file' utility.
      // MIME content type format defined in http://www.ietf.org/rfc/rfc1521.txt
      $output = exec('file -bi ' . escapeshellarg($filename), $ignore, $status);
      if ($status == 0 && preg_match('!([\w-]+/[\w\d_+-]+)!', $output, $matches))
        return $matches[1];
    }

    return 'application/octet-stream';
  }
}

//////////////////////////////////////////////////////////////////////////////
// Repository implementation

/**
 * Implements a file-system-backed Bitcache repository.
 */
class Bitcache_Repository implements Iterator, Countable {
  public $path, $uri;

  function __construct($path, $options = array()) {
    $this->path = $path;
    $this->uri = preg_match('![\w]+://!', $path) ? $path : 'file://' . realpath($path);
  }

  /**
   * Determines how much disk space is occupied by the repository.
   */
  function size() { // FIXME
    $info = explode("\t", shell_exec('du -k ' . $this->path));
    return (int)$info[0] * 1024;
  }

  /**
   * Determines the total bitstream count in the repository.
   */
  function count() {
    $count = 0;
    foreach ($this as $stream) { $count++; }
    return $count;
  }

  function rewind() {
    if (!empty($this->dir)) {
      closedir($this->dir);
    }
    $this->dir = is_dir($this->path) ? opendir($this->path) : NULL;
    $this->next();
  }

  function key() {
    return $this->file;
  }

  function current() {
    return new Bitcache_Stream($this->path($this->file));
  }

  function next() {
    if (!empty($this->dir)) {
      while (($this->file = readdir($this->dir)) !== FALSE) {
        if (preg_match(BITCACHE_ID_FORMAT, $this->file)) {
          break;
        }
      }
    }
  }

  function valid() {
    return !empty($this->dir) && $this->file !== FALSE;
  }

  function exists($id) {
    return is_readable($this->path($id));
  }

  function get($id) {
    if (is_readable($filepath = $this->path($id))) {
      return new Bitcache_Stream($filepath);
    }
  }

  function put($id, $data) {
    $id = !$id ? sha1($data) : $id;
    return file_put_contents($this->path($id), $data) ? $id : FALSE;
  }

  function put_file($id, $filepath, $move = FALSE) {
    $id = !$id ? sha1_file($filepath) : $id;
    return ($move ?
      rename($filepath, $this->path($id)) :
      copy($filepath, $this->path($id))) ? $id : FALSE;
  }

  function delete($id) {
    return @unlink($this->path($id));
  }

  protected function path($id = NULL) {
    return $this->path . ($id ? '/' . $id : '');
  }
}

//////////////////////////////////////////////////////////////////////////////
// REST API server implementation

/**
 * Implements a Bitcache REST API server.
 */
class Bitcache_Server {
  function __construct($repo, $options = array()) {
    $this->repo = is_object($repo) ? $repo : new Bitcache_Repository((string)$repo);
    $this->options = $options;
  }

  /**
   * Handles OPTIONS requests. Determines the set of valid HTTP methods for
   * the bitstream index or a bitstream resource.
   *
   * @param  string $request the request URI
   * @param  array  $query   associative array of query string arguments
   * @link   <http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.2>
   */
  function options($request, $query = array()) {
    $this->header('Content-Type: text/plain; charset=ascii');

    if (!($id = $this->resolve('OPTIONS', $request))) {
      // Options for the bitstream index
      $methods = array_filter(array('OPTIONS', 'GET', 'POST'), array($this, 'allow'));
    }
    else {
      // Options for a bitstream resource
      $methods = array('OPTIONS', 'GET', 'HEAD', 'PUT', 'DELETE');
      foreach ($methods as $key => $method) {
        if (!$this->allow($method, $id))
          unset($methods[$key]);
      }
    }
    $this->header('Accept: ' . implode(', ', $methods));

    return TRUE;
  }

  /**
   * Handles INDEX requests. Outputs the full index of available bitstreams.
   *
   * @param  string $request the request URI
   * @param  array  $query   associative array of query string arguments
   */
  function index($request, $query = array(), $body = TRUE) {
    $this->header('Content-Type: text/plain; charset=ascii');

    if ($body) {
      $this->index_text();
    }

    return TRUE;
  }

  /**
   * Handles HEAD requests.
   *
   * @param  string $request the request URI
   * @param  array  $query   associative array of query string arguments
   * @link   <http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.4>
   */
  function head($request, $query = array()) {
    return $this->get($request, $query, FALSE);
  }

  /**
   * Handles GET requests.
   *
   *
   * @param  string $request the request URI
   * @param  array  $query   associative array of query string arguments
   * @link   <http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.3>
   */
  function get($request, $query = array(), $body = TRUE) {
    if (!($id = $this->resolve('GET', $request)))
      return $this->index($request, $query, $body);

    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.4.5>
    if (!($stream = $this->repo->get($id)))
      return $this->abort(404, 'Not Found');

    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.5.4>
    if (!$stream->open())
      return $this->abort(503, 'Service Unavailable');

    $this->headers($id, $stream);

    if ($body) {
      // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.13>
      $this->header('Content-Length: ' . $stream->size);

      $this->passthru($id, $stream);
    }

    $stream->close();

    return TRUE;
  }

  /**
   * Handles POST requests.
   *
   * @param  string $request the request URI
   * @param  array  $query   associative array of query string arguments
   * @link   <http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.5>
   */
  function post($request, $query = array()) {
    // We only support POST requests to the bitstream index.
      // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.4.1>
    if ($request != '/')
      return $this->abort(400, 'Bad Request');

    $this->upload(NULL);

    return TRUE;
  }

  /**
   * Handles PUT requests.
   *
   * @param  string $request the request URI
   * @param  array  $query   associative array of query string arguments
   * @link   <http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.6>
   */
  function put($request, $query = array()) {
    $this->upload($this->resolve('PUT', $request));

    return TRUE;
  }

  /**
   * Handles DELETE requests.
   *
   * @param  string $request the request URI
   * @param  array  $query   associative array of query string arguments
   * @link   <http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9.7>
   */
  function delete($request, $query = array()) {
    $id = $this->resolve('DELETE', $request);

    if (!$this->repo->delete($id))
      return $this->abort(503, 'Service Unavailable');

    $this->deleted($id);

    return TRUE;
  }

  protected function resolve($method, $request) {
    if ($request[0] != '/')
      return $this->abort(400, 'Bad Request');

    $id = substr($request, 1); // strip off the leading '/'

    // We don't support requests for anything else except the root
    // collection '/' and for bitstreams identified by fingerprints of the
    // correct length and written in all lowercase hexadecimal letters
    // '/[a-z0-9]+'. Note that the reason for returning 400 Bad Request
    // instead of 404 Not Found is to differentiate actually invalid
    // requests from valid requests for an unknown bitstream.
    if (!empty($id) && !preg_match(BITCACHE_ID_FORMAT, $id))
      return $this->abort(400, 'Bad Request');

    if (!$this->allow($method, $id))
      return $this->abort(403, 'Forbidden');

    return $id;
  }

  protected function allow($method, $id = NULL) {
    return TRUE; // Can be overridden in subclasses
  }

  protected function upload($id = NULL) {
    // TODO: $_SERVER['CONTENT_TYPE'] != 'application/x-www-form-urlencoded'

    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.4.12>
    if (!isset($_SERVER['CONTENT_LENGTH']) || !ctype_digit($_SERVER['CONTENT_LENGTH']))
      return $this->abort(411, 'Length Required');

    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.4.14>
    if (BITCACHE_POST_MAX_SIZE && (int)$_SERVER['CONTENT_LENGTH'] > BITCACHE_POST_MAX_SIZE)
      return $this->abort(413, 'Request Entity Too Large');

    $tmpfile = $this->receive();
    $sha1 = sha1_file($tmpfile);

    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.4.10>
    // If the actual fingerprint does not match the fingerprint provided us
    // by the client's PUT request, we are dealing either with in-transit
    // data corruption or a buggy/malicious client. In any case, time to
    // call it quits.
    if (!empty($id) && $sha1 != $id)
      return $this->abort(409, 'Conflict');

    if (!$this->repo->put_file($sha1, $tmpfile, TRUE))
      return $this->abort(503, 'Service Unavailable');

    $this->created($sha1, $stream);

    return TRUE;
  }

  protected function created($id, $stream) {
    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.2.2>
    $this->status(201, 'Created');

    $this->header('Content-Type: application/octet-stream'); // FIXME?
    $this->header('ETag: "' . $id . '"');
    $this->redirect('/' . $id);
  }

  protected function deleted($id) {
    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.2.5>
    $this->status(204, 'No Content');
  }

  protected function abort($code, $msg = '') {
    $this->status($code, $msg);

    $this->header('Content-Type: text/plain; charset=ascii');
    die(trim("$code $msg") . "\n");
  }

  protected function redirect($path = '/') {
    $this->header('Location: ' . $path);
  }

  protected function status($code, $msg = '') {
    $this->header(trim("HTTP/1.1 $code $msg"));
  }

  protected function header($text) {
    header($text);
  }

  protected function headers($id, $stream) {
    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.35>
    $this->header('Accept-Ranges: bytes');

    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.11>
    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.19>
    $this->header('ETag: "' . $id . '"');

    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.15>
    // A Content-SHA1 header is not specified in HTTP/1.1, but it does
    // specify Content-MD5; this is a straightforward, if non-standard,
    // extension of that same concept.
    $this->header('Content-SHA1: ' . $id);

    // <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.17>
    $this->header('Content-Type: ' . $stream->type);
  }

  protected function passthru($id, $stream) {
    fpassthru($stream->stream);
  }

  protected function receive() {
    $input   = fopen('php://input', 'rb');
    $tmpdir  = function_exists('sys_get_temp_dir') ? sys_get_temp_dir() : '/tmp';
    $tmpfile = tempnam($tmpdir, 'bitcache');
    $output  = fopen($tmpfile, 'wb');
    stream_copy_to_stream($input, $output);
    fclose($input);
    fclose($output);
    return $tmpfile;
  }

  protected function index_text() {
    foreach ($this->repo as $id => $stream) {
      printf("%s\t%d\n", $id, $stream->size);
    }
  }
}
